'#################################################
'# 
'# fbList: Fast arbitrary index insert/remove simple extensible list with 
'#          support for strongly typed lists including user defined types
'# 
'# Version: 0.3
'# 
'# Official site: www.JatteGames.com
'# 
'# fbList is Copyright (c) 2023-2024 Johannes "Jattenalle" Pihl, all rights reserved
'# 
'# 	This software is provided 'as-is', without any express or implied
'# 	warranty. In no event will the authors be held liable for any damages
'# 	arising from the use of this software.
'# 	
'# 	Permission is granted to anyone to use this software for any purpose,
'# 	including commercial applications, and to alter it and redistribute it
'# 	freely, subject to the following restrictions:
'# 	
'# 	1. The origin of this software must not be misrepresented; you must not
'# 	   claim that you wrote the original software. If you use this software
'# 	   in a product, an acknowledgment in the product documentation would be
'# 	   appreciated but is not required.
'# 	2. Altered source versions must be plainly marked as such, and must not be
'# 	   misrepresented as being the original software.
'# 	3. This notice may not be removed or altered from any source distribution.
'# 
'#################################################
'# 
'# Known issues:
'#   List operations are not atomic, some care must be taken if accessing the
'#   same list from multiple different threads at the same time.
'# 
'#################################################
'# 
'# Custom compare function structure:
'#   function(byval A as any ptr, byval B as any ptr) as integer
'# 
'#################################################
'# 
'# Use #DEFINE FBLIST_INCLUDE_HEADER_ONLY before you include fbList to only
'# include the declares and definitions without the methods and procedures
'# themselves.
'# Example:
'#   #DEFINE FBLIST_INCLUDE_HEADER_ONLY
'#   #INCLUDE ONCE "fbList.bi"
'# 
'#################################################
'# 
'# Use #define __FBLIST_PP_DISABLEMACROS to disable TYPELIST, HOLDS,
'# EXTENSIONBEGIN, and EXTENSIONEND, -macros
'# 
'#################################################
'# 
'# Use #define __FBLIST_PP_NOFOREACH to disable the .ForEach macro
'# 
'#################################################

#ifdef FBLIST_INC
	#cmdline "-maxerr 1"
	#error  "FBLIST is already included! It can only be included once to ensure #defines are handled properly. Please verify that you only #include FBLIST once."
#else
	#define FBLIST_INC

	#define FBLIST_VERSION			"0.3"
	
	#if __FB_VERSION__ < "1.10.1"
		#cmdline "-maxerr 1"
		#print "You are using FBC version "
		#print __FB_VERSION__
		#print "FBLIST requires FreeBASIC 1.10.1 or later!"
		#error "Update FreeBASIC!"
	#endif
	#ifndef __FBLIST_PP_NOFOREACH
		#ifdef ForEach
			#cmdline "-maxerr 1"
			#error "FBLIST: FOREACH() is already defined! List.FOREACH() not available! Consider using #define __FBLIST_PP_NOFOREACH"
		#else
			#macro ForEach?(_a)
				__A_List_Is_Required_Before_ForEach()
				for _a = 0 to ___T_LIST_FOREACHTARGET
			#endmacro
		#endif
	#endif
	
	#include once "crt/string.bi"
	#include once "fbc-int/symbol.bi"
	
	const FBLIST_ERR_OUTOFMEM	=	4
	
	#define __T_LIST_CSVNAME		CSV
	#define __T_LIST_COMPARENAME	Compare
	#define __T_LIST_COPYNAME		Copy
	
	const LIST_FIRST	=	0
	const LIST_LAST		=	-1

	const ___T_LIST_ENTRYSIZE = sizeof(any ptr)
	dim shared as integer ___T_LIST_FOREACHTARGET

	declare sub __FBLIST_ERROR(byref msg as string)
	#ifndef FBLIST_INCLUDE_HEADER_ONLY
		sub __FBLIST_ERROR(byref msg as string)
			dim as integer f = freefile
			open err for output as #f
			print #f, msg
			close #f
			open cons for output as #f
			print #f, msg
			close #f
		end sub
	#endif
	
	declare function FBLIST_ODDSEQUENCE(byval index as integer, byval count as integer) as integer
	declare function FBLIST_EVENSEQUENCE(byval index as integer, byval count as integer) as integer
	declare function FBLIST_POTSEQUENCE(byval index as integer, byval count as integer) as integer
	declare function FBLIST_INTSEQUENCE(byval index as integer, byval count as integer) as integer
	declare function __FBLIST_t_List_PFNDEFAULTCOMPARENAME(byval _a as any ptr, byval _b as any ptr) as integer
	#ifndef FBLIST_INCLUDE_HEADER_ONLY
		function FBLIST_EVENSEQUENCE(byval index as integer, byval count as integer) as integer
			return (index + 1) * 2
		end function
		function FBLIST_ODDSEQUENCE(byval index as integer, byval count as integer) as integer
			return (index * 2) + 1
		end function
		function FBLIST_POTSEQUENCE(byval index as integer, byval count as integer) as integer
			return (2 ^ index)
		end function
		function FBLIST_INTSEQUENCE(byval index as integer, byval count as integer) as integer
			return (index + 1)
		end function
		function __FBLIST_t_List_PFNDEFAULTCOMPARENAME(byval _a as any ptr, byval _b as any ptr) as integer
			if _a < _b then
				return -1
			elseif _a > _b then
				return 1
			end if
			return 0
		end function
	#endif
	type __T_LIST_FOUNDATION
		'// Core methods used by all derived list types
		declare function __Merge(byval _p as any ptr, byval index as integer) as integer
		declare function __Replace(byval _p as any ptr, byval index as integer) as integer
		declare function __SortedInsert(byval _p as any ptr) as integer
		declare function __Insert(byval _p as any ptr, byval index as integer, byval _sortparam as integer) as integer
		declare function __Find(byval _match as any ptr, byval _matchIndex as integer = 0) as integer
		declare function __ParseIndex(byval index as integer) as integer
		declare sub __eldtor(byval index as integer)
		declare sub __A_List_Is_Required_Before_ForEach()
		
		'// These methods don't care about the type
		declare Constructor()
		declare destructor()
		declare sub Free()
		
		declare sub Clear()
		
		declare function Shift(byval _first as integer, byval _entries as integer, byval toIndex as integer) as integer
		
		declare sub Remove(byval index as integer)
		
		declare function __T_LIST_CSVNAME() as string
		
		declare sub DefaultCompare()
		declare sub CustomCompare(byval pfnCompare as any ptr)
		
		declare sub AllowDuplicates()
		declare sub RemoveDuplicates()
		declare sub DisallowDuplicates()
		
		declare sub CustomSort(byval pfnCompare as any ptr = 0)
		declare sub EnableSort()
		declare sub Sort()
		declare function IsSorted() as integer
		declare sub DisableSort()
		declare sub Shuffle()
		declare sub Switch(byval a as integer, byval b as integer)
		
		declare sub Alloc(byval entries as integer)
		declare function Length() as integer
		declare function AllocSize() as integer
		
		declare sub EnableCopy()
		declare sub DisableCopy()
		
		declare sub EnableDestructor()
		declare sub DisableDestructor()
		
		RawListMemory as any ptr ptr
		
		union
			Flags as integer
			type
				_UseDestructor: 1 as integer
				_UseCopy: 1 as integer
				_CallDestructor: 1 as integer
				_AllowDuplicate: 1 as integer
				_ForceSorted: 1 as integer
				_IsSorted: 1 as integer
				__Reserved: 25 as integer
			end type
		end union
		_Length as integer
		_AllocSize as integer
		__pfnCompare as function(byval _a as any ptr, byval _b as any ptr) as integer
	end type
	#ifndef FBLIST_INCLUDE_HEADER_ONLY
		destructor __T_LIST_FOUNDATION()
			this.Free()
		end destructor
		
		Constructor __T_LIST_FOUNDATION()
			'// Defaults
			this.__pfnCompare		=	@__FBLIST_t_List_PFNDEFAULTCOMPARENAME
			this._UseDestructor		=	TRUE
			this._UseCopy			=	TRUE
			this._CallDestructor	=	TRUE
			this._AllowDuplicate	=	TRUE
			this._ForceSorted		=	FALSE
			this._IsSorted			=	TRUE
		end Constructor
		
		sub __T_LIST_FOUNDATION.Free()
			this.Clear()
			deallocate this.RawListMemory
			this.RawListMemory	=	0
			this._AllocSize		=	0
		end sub
		
		sub __T_LIST_FOUNDATION.Clear()
			for i as integer = 0 to this._Length - 1
				this.__eldtor(i)
			next
			this._Length = 0
		end sub
		
		
		
		function __T_LIST_FOUNDATION.__Merge(byval _p as any ptr, byval index as integer) as integer
			index = this.__ParseIndex(index)
			if index = this._Length then
				index = this.__Insert(_p, index, FALSE)
			else
				this.__eldtor(index)
				This.RawListMemory[index]	=	_p
			end if
			this._IsSorted		=	FALSE
			return index
		end function
		
		function __T_LIST_FOUNDATION.__Replace(byval _p as any ptr, byval index as integer) as integer
			index = this.__ParseIndex(index)
			if index < this._Length then
				this.__eldtor(index)
				this.RawListMemory[index]	=	_p
				this._IsSorted				=	FALSE
			end if
			return index
		end function
		
		function __T_LIST_FOUNDATION.__SortedInsert(byval _p as any ptr) as integer
			if this._IsSorted = FALSE then '// Sort the existing list
				this.Sort()
			end if
			
			if this._Length then
				dim as integer _length	=	This._Length - 1
				dim as integer pivot	=	_length / 2
				dim as integer min		=	0
				dim as integer max		=	_length
				
				dim as integer compareResult
				
				do while _length > 0
					compareResult = this.__pfnCompare(_p, This.RawListMemory[pivot])
					if compareResult < 0 then '// Less
						max		=	pivot - 1
						_length	=	max - min
						pivot	=	min + _length / 2
					elseif compareResult > 0 then
						min		=	pivot + 1
						_length	=	max - min
						pivot	=	min + _length / 2
					else '// Duplicate
						if this._AllowDuplicate then
							return this.__Insert(_p, pivot, TRUE)
						end if
						return pivot
					end if
				loop
				compareResult = this.__pfnCompare(_p, This.RawListMemory[pivot])
				if compareResult > 0 then
					pivot += 1
				elseif compareResult = 0 then '// Duplicate
					if this._AllowDuplicate then
						return this.__Insert(_p, pivot, TRUE)
					end if
					return pivot
				end if
				if pivot < 0 then
					pivot = 0
				end if
				
				return this.__Insert(_p, pivot, TRUE)
			else '// Just insert it, there's nothing else to do
				return this.__Insert(_p, LIST_LAST, TRUE)
			end if
			return -1
		end function
		
		function __T_LIST_FOUNDATION.__Insert(byval _p as any ptr, byval index as integer, byval _sortparam as integer) as integer
			index = this.__ParseIndex(index)
			
			this._IsSorted		=	_sortparam
			
			this.Alloc(This._Length + 1)
			
			if this._AllowDuplicate = FALSE then
				dim as integer r = this.__Find(_p)
				if r > -1 then
					return r
				end if
			end if
			
			if index < This._Length then '// Anywhere except adding to the end
				if This._Length + 1 > This._AllocSize then '// Grow it, and split it while we're at it
					This._AllocSize	+=	1
				end if
				dim as any ptr _hold = This.RawListMemory
				This.RawListMemory		=	callocate(This._AllocSize * ___T_LIST_ENTRYSIZE)
				if This.RawListMemory = 0 then
					__FBLIST_ERROR("ERROR: FBLIST __Insert failed to allocate "& This._AllocSize * ___T_LIST_ENTRYSIZE &" bytes!")
					error FBLIST_ERR_OUTOFMEM '// Out of memory
				end if
				if index > 0 then
					memcpy(This.RawListMemory, _hold, index * ___T_LIST_ENTRYSIZE)
				end if
				memcpy(This.RawListMemory + (index + 1), _hold + (index * ___T_LIST_ENTRYSIZE), (This._Length-index) * ___T_LIST_ENTRYSIZE)
				deallocate _hold
				
				This.RawListMemory[index]	=	cptr(any ptr, _p)
				This._Length			+=	1
				
				return index
			else
				This.RawListMemory[This._Length]	=	cptr(any ptr, _p)
				This._Length				+=	1
				return This._Length - 1
			end if
			return -1
		end function
		
		function __T_LIST_FOUNDATION.__Find(byval _match as any ptr, byval _matchIndex as integer = 0) as integer
			if this._Length then
				dim as integer _matchCount
				dim as integer compareResult
				if this._IsSorted then
					dim as integer _length	=	This._Length - 1
					dim as integer pivot	=	_length / 2
					dim as integer min		=	0
					dim as integer max		=	_length
					
					do while _length > 0
						if This.RawListMemory[pivot] = _match then
							if _matchCount = _matchIndex then
								return pivot
							end if
							_matchCount += 1
						else
							compareResult = this.__pfnCompare(_match, This.RawListMemory[pivot])
							if compareResult < 0 then '// Less
								max		=	pivot - 1
								_length	=	max - min
								pivot	=	min + _length / 2
							elseif compareResult > 0 then
								min		=	pivot + 1
								_length	=	max - min
								pivot	=	min + _length / 2
							else '// Duplicate
								if _matchCount = _matchIndex then
									return pivot
								end if
								_matchCount += 1
							end if
						end if
					loop
					if This.RawListMemory[pivot] = _match then
						if _matchCount = _matchIndex then
							return pivot
						end if
						_matchCount += 1
					end if
					compareResult = this.__pfnCompare(_match, This.RawListMemory[pivot])
					if compareResult = 0 then '// Duplicate
						if _matchCount = _matchIndex then
							return pivot
						end if
						_matchCount += 1
					end if
				else
					for p as integer = 0 to this._Length - 1
						if This.RawListMemory[p] = _match then
							if _matchCount = _matchIndex then
								return p
							end if
							_matchCount += 1
						else
							compareResult = this.__pfnCompare(This.RawListMemory[p], _match)
							if compareResult = 0 then
								if _matchCount = _matchIndex then
									return p
								end if
								_matchCount += 1
							end if
						end if
					next
				end if
			end if
			return -1
		end function
		
		
		
		
		
		
		
		
		
		
		sub __T_LIST_FOUNDATION.Remove(byval index as integer)
			if index = LIST_LAST then
				index = This._Length - 1
			end if
			if index < 0 or index >= This._Length then
				exit sub
			end if
			this.__eldtor(index)
			memmove(This.RawListMemory + index, This.RawListMemory + index +1, (This._Length-index-1) * ___T_LIST_ENTRYSIZE)
			This.RawListMemory[This._Length - 1] = 0
			This._Length				-=	1
		end sub
		
		sub __T_LIST_FOUNDATION.RemoveDuplicates()
			dim as integer compareResult
			if this._IsSorted then
				for p as integer = 0 to this._Length - 2
					if p >= this._Length then '// We must re-check as elements may have been removed
						exit for
					end if
					for c as integer = p + 1 to this._Length - 1
						if c >= this._Length then '// We must re-check as elements may have been removed
							exit for
						end if
						if This.RawListMemory[p] = This.RawListMemory[c] then
							this.Remove(c)
							c -= 1
						else
							compareResult = this.__pfnCompare(This.RawListMemory[p], This.RawListMemory[c])
							if compareResult = 0 then
								this.Remove(c)
								c -= 1
							else
								exit for '// The list is sorted, so if the values don't match then there can be no more duplicates
							end if
						end if
					next
				next
			else
				for p as integer = 0 to this._Length - 1
					if p >= this._Length then '// We must re-check as elements may have been removed
						exit for
					end if
					for c as integer = p+1 to this._Length - 1
						if c >= this._Length then '// We must re-check as elements may have been removed
							exit for
						end if
						if This.RawListMemory[p] = This.RawListMemory[c] then
							this.Remove(c)
							c -= 1
						else
							compareResult = this.__pfnCompare(This.RawListMemory[p], This.RawListMemory[c])
							if compareResult = 0 then
								this.Remove(c)
								c -= 1
							end if
						end if
					next
				next
			end if
		end sub
		
		function __T_LIST_FOUNDATION.Shift(byval _first as integer, byval _entries as integer, byval toIndex as integer = LIST_LAST) as integer
			_first = this.__ParseIndex(_first)
			
			if _entries = LIST_LAST then
				_entries = this._Length - _first
			end if
			
			if toIndex = LIST_LAST then
				toIndex = this._Length - _entries
			end if
			toIndex = this.__ParseIndex(toIndex)
			
			if _entries > 0 then
				if toIndex + _entries <= this._Length then
					if _first + _entries <= this._Length then
						this._IsSorted = FALSE
						dim as any ptr _hold = callocate(_entries * ___T_LIST_ENTRYSIZE)
						if _hold = 0 then
							__FBLIST_ERROR("ERROR: FBLIST Shift failed to allocate "& _entries * ___T_LIST_ENTRYSIZE &" bytes!")
							error FBLIST_ERR_OUTOFMEM '// Out of memory
						end if
						memcpy(_hold, This.RawListMemory + _first, _entries * ___T_LIST_ENTRYSIZE)
						memmove(This.RawListMemory + _first, This.RawListMemory + (_first + _entries), (this._Length - _first - _entries) * ___T_LIST_ENTRYSIZE)
						memmove(This.RawListMemory + toIndex + _entries, This.RawListMemory + toIndex, (this._Length - toIndex -_entries) * ___T_LIST_ENTRYSIZE)
						memcpy(This.RawListMemory + toIndex, _hold, _entries * ___T_LIST_ENTRYSIZE)
						deallocate _hold
					end if
				end if
			end if
			return toIndex
		end function
		
		sub __T_LIST_FOUNDATION.DefaultCompare()
			this.CustomCompare(@__FBLIST_t_List_PFNDEFAULTCOMPARENAME)
		end sub
		
		sub __T_LIST_FOUNDATION.CustomCompare(byval pfnCompare as any ptr)
			if pfnCompare <> this.__pfnCompare then
				this._IsSorted		=	FALSE
			end if
			if pfnCompare then
				this.__pfnCompare = pfnCompare
			end if
		end sub
		
		sub __T_LIST_FOUNDATION.CustomSort(byval pfnCompare as any ptr)
			this.CustomCompare(pfnCompare)
		end sub
		
		sub __T_LIST_FOUNDATION.EnableSort()
			this._ForceSorted = TRUE
		end sub
		
		sub __T_LIST_FOUNDATION.Sort()
			this._IsSorted		=	TRUE
			if this._Length < 2 then '// There's either no entries, or just one entry, in the list so we don't need to sort it
				exit sub
			else
				dim as integer compareResult
				for p as integer = 0 to this._Length - 2
					for c as integer = p+1 to this._Length - 1
						compareResult = this.__pfnCompare(This.RawListMemory[p], This.RawListMemory[c])
						if compareResult = 0 then '// Duplicate
							
						elseif compareResult > 0 then
							this.Switch(p, c)
						end if
					next
				next
			end if
		end sub
		
		
		
		
		
		sub __T_LIST_FOUNDATION.Alloc(byval entries as integer)
			if entries > this._AllocSize then
				dim as integer addSize = int(this._Length * 0.1) '// Add 10% of the current list size
				if addSize > 8192 then '// Don't add more than this extra, even if we could
					addSize = 8192
				end if
				entries += addSize
				
				dim as integer mSize = entries * ___T_LIST_ENTRYSIZE
				this._AllocSize = entries
				
				if this._Length = 0 then
					if RawListMemory then
						deallocate this.RawListMemory
					end if
					this.RawListMemory = callocate(mSize)
					if This.RawListMemory = 0 then
						__FBLIST_ERROR("ERROR: FBLIST Alloc("& mSize &") failed! Out of memory!")
						error FBLIST_ERR_OUTOFMEM '// Out of memory
					end if
					this._AllocSize = entries
				else
					dim as any ptr _hold = This.RawListMemory
					This.RawListMemory		=	callocate(mSize)
					if This.RawListMemory = 0 then
						__FBLIST_ERROR("ERROR: FBLIST Alloc("& mSize &") failed! Out of memory!")
						error FBLIST_ERR_OUTOFMEM '// Out of memory
					end if
					if _hold then
						memcpy(This.RawListMemory, _hold, this._Length * ___T_LIST_ENTRYSIZE)
						deallocate _hold
					end if
				end if
			end if
		end sub
		
		sub __T_LIST_FOUNDATION.Switch(byval a as integer, byval b as integer)
			if a = LIST_LAST then
				a = This._Length - 1
			end if
			if a < 0 or a >= This._Length then
				exit sub
			end if
			if b = LIST_LAST then
				b = This._Length - 1
			end if
			if b < 0 or b >= This._Length then
				exit sub
			end if
			if a <> b then
				swap This.RawListMemory[a], This.RawListMemory[b]
				this._IsSorted		=	FALSE
			end if
		end sub
		
		function __T_LIST_FOUNDATION.IsSorted() as integer
			return this._IsSorted
		end function
		
		function __T_LIST_FOUNDATION.Length() as integer
			return this._Length
		end function
		
		function __T_LIST_FOUNDATION.AllocSize() as integer
			return this._AllocSize
		end function
		
		sub __T_LIST_FOUNDATION.DisallowDuplicates()
			this._AllowDuplicate	=	FALSE
		end sub
		
		sub __T_LIST_FOUNDATION.AllowDuplicates()
			this._AllowDuplicate	=	TRUE
		end sub
		
		sub __T_LIST_FOUNDATION.DisableSort()
			this._ForceSorted	=	FALSE
		end sub
		
		sub __T_LIST_FOUNDATION.EnableCopy()
			this._UseCopy	=	TRUE
		end sub
		
		sub __T_LIST_FOUNDATION.DisableCopy()
			this._UseCopy	=	FALSE
		end sub
		
		sub __T_LIST_FOUNDATION.EnableDestructor()
			this._UseDestructor	=	TRUE
		end sub
		
		sub __T_LIST_FOUNDATION.DisableDestructor()
			this._UseDestructor	=	FALSE
		end sub
		
		
		sub __T_LIST_FOUNDATION.Shuffle()
			if this._Length > 1 then '// Can't shuffle less than 2 elements
				this._IsSorted	=	FALSE
				dim as integer i = this._Length - 1
				dim as integer r
				do
					r = int((i+1) * rnd)
					this.Switch(i, r)
					i -= 1
				loop while i > 0
			end if
		end sub
		
		function __T_LIST_FOUNDATION.__T_LIST_CSVNAME() as string
			dim as string r
			if this._Length then
				for i as integer = 0 to this._Length - 1
					r &= this.RawListMemory[i]
					if i < this._Length -1 then
						r &=  ", "
					end if
				next
			end if
			return r
		end function
		
		function __T_LIST_FOUNDATION.__ParseIndex(byval index as integer) as integer
			if index = LIST_LAST then
				index = this.Length
			end if
			if index < 0 then
				index = 0
			end if
			if index > this.Length then
				index = this.Length
			end if
			return index
		end function
		
		sub __T_LIST_FOUNDATION.__A_List_Is_Required_Before_ForEach()
			___T_LIST_FOREACHTARGET = this._Length - 1
		end sub
		
		sub __T_LIST_FOUNDATION.__eldtor(byval index as integer)
			if index >= 0 and index < this._AllocSize then
				this.RawListMemory[index] = 0
			end if
		end sub
		
	#endif

	type t_List extends __T_LIST_FOUNDATION
		declare function Fill(byref _l as t_List, byval _len as integer, byval index as integer = LIST_LAST) as integer
		declare function Fill(byval _l as t_List ptr, byval _len as integer, byval index as integer = LIST_LAST) as integer
		declare function Fill(_v() as integer, byval _len as integer, byval index as integer = LIST_LAST) as integer
		declare function Fill(_v() as any ptr, byval _len as integer, byval index as integer = LIST_LAST) as integer
		declare function Fill(byref _p as integer, byval _len as integer, byval index as integer = LIST_LAST) as integer
		declare function Fill(byval _p as any ptr, byval _len as integer, byval index as integer = LIST_LAST) as integer
		
		declare function IntSequence(byval _len as integer, byval index as integer = LIST_LAST) as integer
		declare function PoTSequence(byval _len as integer, byval index as integer = LIST_LAST) as integer
		declare function OddSequence(byval _len as integer, byval index as integer = LIST_LAST) as integer
		declare function EvenSequence(byval _len as integer, byval index as integer = LIST_LAST) as integer
		declare function Sequence(byval pfnGenerator as function(byval listIndex as integer, byval sequenceCount as integer) as integer, byval _len as integer, byval index as integer = LIST_LAST) as integer
		
		declare function Merge(byref _l as t_List, byval index as integer) as integer
		declare function Merge(byval _l as t_List ptr, byval index as integer) as integer
		declare function Merge(_v() as integer, byval index as integer) as integer
		declare function Merge(_v() as any ptr, byval index as integer) as integer
		declare function Merge(byref _p as integer, byval index as integer) as integer
		declare function Merge(byval _p as any ptr, byval index as integer) as integer
		
		declare function Replace(byref _l as t_List, byval index as integer) as integer
		declare function Replace(byval _l as t_List ptr, byval index as integer) as integer
		declare function Replace(_v() as integer, byval index as integer) as integer
		declare function Replace(_v() as any ptr, byval index as integer) as integer
		declare function Replace(byref _p as integer, byval index as integer) as integer
		declare function Replace(byval _p as any ptr, byval index as integer) as integer
		
		declare function Insert(byref _l as t_List, byval index as integer = LIST_LAST) as integer
		declare function Insert(byval _l as t_List ptr, byval index as integer = LIST_LAST) as integer
		declare function Insert(_v() as integer, byval index as integer = LIST_LAST) as integer
		declare function Insert(_v() as any ptr, byval index as integer = LIST_LAST) as integer
		declare function Insert(byref _p as integer, byval index as integer = LIST_LAST) as integer
		declare function Insert(byval _p as any ptr, byval index as integer = LIST_LAST) as integer
		
		declare operator +=(byref _l as t_List)
		declare operator +=(byval _l as t_List ptr)
		declare operator +=(byref _p as integer)
		declare operator +=(byval _p as any ptr)
		
		declare operator Let(byref _l as t_List)
		declare operator Let(byval _l as t_List ptr)
		declare operator Let(byref _p as integer)
		declare operator Let(byval _p as any ptr)
		
		declare function SortedInsert(byref _l as t_List) as integer
		declare function SortedInsert(byval _l as t_List ptr) as integer
		declare function SortedInsert(_v() as integer) as integer
		declare function SortedInsert(_v() as any ptr) as integer
		declare function SortedInsert(byref _p as integer) as integer
		declare function SortedInsert(byval _p as any ptr) as integer
		
		declare Constructor (byref _l as t_List)
		declare Constructor (byval _l as t_List ptr)
		declare Constructor (_v() as integer)
		declare Constructor (_v() as any ptr)
		declare Constructor (byref _p as integer)
		declare Constructor (byval _p as any ptr)
		declare Constructor ()
		
		declare sub Clear(byref _l as t_List)
		declare sub Clear(byval _l as t_List ptr)
		declare sub Clear(_v() as integer)
		declare sub Clear(_v() as any ptr)
		declare sub Clear(byref _p as integer)
		declare sub Clear(byval _p as any ptr)
		declare sub Clear()
		
		declare function Find(byval _match as integer, byval _matchIndex as integer = 0) as integer
		declare function Find(byval _match as any ptr, byval _matchIndex as integer = 0) as integer
		
		declare function GetPtr(byval index as integer) as any ptr
		
		declare operator [] (byval index as integer) byref as integer
		
		declare function __Copy(byref _in as integer) as any ptr
	end type
	#ifndef FBLIST_INCLUDE_HEADER_ONLY
		Constructor t_List(byref _l as t_List)
			this.Constructor()
			this.Insert(_l)
		end constructor
		Constructor t_List(byval _l as t_List ptr)
			this.Constructor()
			this.Insert(_l)
		end constructor
		Constructor t_List(_v() as integer)
			this.Constructor()
			this.Insert(_v())
		end constructor
		Constructor t_List(_v() as any ptr)
			this.Constructor()
			this.Insert(_v())
		end constructor
		Constructor t_List(byref _p as integer)
			this.Constructor()
			this.Insert(_p)
		end constructor
		Constructor t_List(byval _p as any ptr)
			this.Constructor()
			this.Insert(_p)
		end constructor
		Constructor t_List()
			base.constructor()
			this.__pfnCompare	=	@__FBLIST_t_List_PFNDEFAULTCOMPARENAME
		end Constructor
		
		
		
		
		
		operator t_List.[] (byval index as integer) byref as integer
			index = this.__ParseIndex(index)
			return cast(integer, this.RawListMemory[index])
		end operator
		function t_List.GetPtr(byval index as integer) as any ptr
			index = this.__ParseIndex(index)
			return this.RawListMemory[index]
		end function
		
		
		
		function t_List.Fill(byref _l as t_List, byval _len as integer, byval index as integer) as integer
			return this.Fill(@_l, _len, index)
		end function
		function t_List.Fill(byval _l as t_List ptr, byval _len as integer, byval index as integer) as integer
			dim as integer killCopy
			if @this = _l then
				_l = new t_List
				*_l = this
				killCopy = TRUE
			end if
			if _l->_Length > 0 then
				index = this.__ParseIndex(index)
				for g as integer = 0 to _len -1
					index = this.Insert(_l->RawListMemory[(g mod _l->_Length)], index) + 1
				next
			end if
			if killCopy then
				_l->Destructor
			end if
			return index
		end function
		function t_List.Fill(_v() as integer, byval _len as integer, byval index as integer) as integer
			index = this.__ParseIndex(index)
			dim as integer lb = lbound(_v)
			dim as integer al = ubound(_v) - lb + 1
			for g as integer = 0 to _len -1
				index = this.Insert(_v(lb + (g mod al)), index) + 1
			next
			return index
		end function
		function t_List.Fill(_v() as any ptr, byval _len as integer, byval index as integer) as integer
			index = this.__ParseIndex(index)
			dim as integer lb = lbound(_v)
			dim as integer al = ubound(_v) - lb + 1
			for g as integer = 0 to _len -1
				index = this.Insert(_v(lb + (g mod al)), index) + 1
			next
			return index
		end function
		function t_List.Fill(byref _p as integer, byval _len as integer, byval index as integer) as integer
			index = this.__ParseIndex(index)
			for g as integer = 0 to _len -1
				index = this.Insert(_p, index) + 1
			next
			return index
		end function
		function t_List.Fill(byval _p as any ptr, byval _len as integer, byval index as integer) as integer
			index = this.__ParseIndex(index)
			for g as integer = 0 to _len -1
				index = this.Insert(_p, index) + 1
			next
			return index
		end function
		
		
		
		
		function t_List.IntSequence(byval _len as integer, byval index as integer) as integer
			return this.Sequence(@FBLIST_INTSEQUENCE, _len, index)
		end function
		function t_List.PoTSequence(byval _len as integer, byval index as integer) as integer
			return this.Sequence(@FBLIST_POTSEQUENCE, _len, index)
		end function
		function t_List.EvenSequence(byval _len as integer, byval index as integer) as integer
			return this.Sequence(@FBLIST_EVENSEQUENCE, _len, index)
		end function
		function t_List.OddSequence(byval _len as integer, byval index as integer) as integer
			return this.Sequence(@FBLIST_ODDSEQUENCE, _len, index)
		end function
		function t_List.Sequence(byval pfnGenerator as function(byval listIndex as integer, byval sequenceCount as integer) as integer, byval _len as integer, byval index as integer) as integer
			if pfnGenerator then
				index = this.__ParseIndex(index)
				for g as integer = 0 to _len -1
					index = this.Insert(cptr(any ptr, pfnGenerator(index, g)), index) + 1
				next
			end if
			return index
		end function
		
		
		
		
		sub t_List.Clear(byref _l as t_List)
			this.Clear()
			this.Insert(_l)
		end sub
		sub t_List.Clear(byval _l as t_List ptr)
			this.Clear()
			this.Insert(_l)
		end sub
		sub t_List.Clear(_v() as integer)
			this.Clear()
			this.Insert(_v())
		end sub
		sub t_List.Clear(_v() as any ptr)
			this.Clear()
			this.Insert(_v())
		end sub
		sub t_List.Clear(byref _p as integer)
			this.Clear()
			this.Insert(_p)
		end sub
		sub t_List.Clear(byval _p as any ptr)
			this.Clear()
			this.Insert(_p)
		end sub
		sub t_List.Clear()
			base.Clear()
		end sub
		
		
		
		
		
		operator t_List.Let(byref _l as t_List)
			this.Clear(_l)
		end operator
		operator t_List.Let(byval _l as t_List ptr)
			this.Clear(_l)
		end operator
		operator t_List.Let(byref _p as integer)
			this.Clear(_p)
		end operator
		operator t_List.Let(byval _p as any ptr)
			this.Clear(_p)
		end operator
		
		
		
		
		
		function t_List.Merge(byref _l as t_List, byval index as integer) as integer
			return this.Merge(@_l, index)
		end function
		function t_List.Merge(byval _l as t_List ptr, byval index as integer) as integer
			dim as integer killCopy
			if @this = _l then
				_l = new t_List
				*_l = this
				killCopy = TRUE
			end if
			dim as integer ni = this.__ParseIndex(index)
			dim as integer ri = -1
			for g as integer = 0 to _l->_Length - 1
				ri = this.Merge( _l->RawListMemory[g], ni)
				if ri > -1 then
					ni += 1
				else
					exit for
				end if
			next
			if killCopy then
				_l->destructor()
			end if
			return ri
		end function
		function t_List.Merge(_v() as integer, byval index as integer) as integer
			dim as integer ni = this.__ParseIndex(index)
			dim as integer ri = -1
			for g as integer = lbound(_v) to ubound(_v)
				ri = this.Merge( _v(g), ni)
				if ri > -1 then
					ni += 1
				else
					exit for
				end if
			next
			return ri
		end function
		function t_List.Merge(_v() as any ptr, byval index as integer) as integer
			dim as integer ni = this.__ParseIndex(index)
			dim as integer ri = -1
			for g as integer = lbound(_v) to ubound(_v)
				ri = this.Merge( _v(g), ni)
				if ri > -1 then
					ni += 1
				else
					exit for
				end if
			next
			return ri
		end function
		function t_List.Merge(byref _p as integer, byval index as integer) as integer
			return this.Merge(cptr(any ptr, _p), index)
		end function
		function t_List.Merge(byval _p as any ptr, byval index as integer) as integer
			return this.__Merge(_p, index)
		end function
		
		
		
		
		
		function t_List.Replace(byref _l as t_List, byval index as integer) as integer
			return this.Replace(@_l, index)
		end function
		function t_List.Replace(byval _l as t_List ptr, byval index as integer) as integer
			dim as integer killCopy
			if @this = _l then
				_l = new t_List
				*_l = this
				killCopy = TRUE
			end if
			dim as integer ni = this.__ParseIndex(index)
			dim as integer ri = -1
			for g as integer = 0 to _l->_Length - 1
				if ni < this._Length then
					ri = this.Replace( _l->RawListMemory[g], ni)
					if ri > -1 then
						ni += 1
					else
						exit for
					end if
				else
					exit for
				end if
			next
			if killCopy then
				_l->destructor()
			end if
			return ri
		end function
		function t_List.Replace(_v() as integer, byval index as integer) as integer
			dim as integer ni = this.__ParseIndex(index)
			dim as integer ri = -1
			for g as integer = lbound(_v) to ubound(_v)
				if ni < this._Length then
					ri = this.Replace( _v(g), ni)
					if ri > -1 then
						ni += 1
					else
						exit for
					end if
				else
					exit for
				end if
			next
			return ri
		end function
		function t_List.Replace(_v() as any ptr, byval index as integer) as integer
			dim as integer ni = this.__ParseIndex(index)
			dim as integer ri = -1
			for g as integer = lbound(_v) to ubound(_v)
				if ni < this._Length then
					ri = this.Replace( _v(g), ni)
					if ri > -1 then
						ni += 1
					else
						exit for
					end if
				else
					exit for
				end if
			next
			return ri
		end function
		function t_List.Replace(byref _p as integer, byval index as integer) as integer
			return this.Replace(cptr(any ptr, _p), index)
		end function
		function t_List.Replace(byval _p as any ptr, byval index as integer) as integer
			return this.__Replace(_p, index)
		end function
		
		
		
		
		
		operator t_List.+=(byref _l as t_List)
			this.Insert(_l)
		end operator
		operator t_List.+=(byval _l as t_List ptr)
			this.Insert(_l)
		end operator
		operator t_List.+=(byref _p as integer)
			this.Insert(_p)
		end operator
		operator t_List.+=(byval _p as any ptr)
			this.Insert(_p)
		end operator
		
		
		
		
		
		function t_List.Find(byval _match as integer, byval _matchIndex as integer = 0) as integer
			return this.__Find(cptr(any ptr, _match), _matchIndex)
		end function
		function t_List.Find(byval _match as any ptr, byval _matchIndex as integer = 0) as integer
			return this.__Find(_match, _matchIndex)
		end function
		
		
		
		
		
		function t_List.Insert(byref _l as t_List, byval index as integer = LIST_LAST) as integer
			return this.Insert(@_l, index)
		end function
		function t_List.Insert(byval _l as t_List ptr, byval index as integer = LIST_LAST) as integer
			dim as integer killCopy
			if @this = _l then
				_l = new t_List
				*_l = this
				killCopy = TRUE
			end if
			index = this.__ParseIndex(index)
			this.Alloc(this._Length + _l->_Length)
			for i as integer = 0 to _l->_Length - 1
				index = this.Insert(_l->RawListMemory[i], index) + 1
			next
			if killCopy then
				_l->destructor()
			end if
			return index
		end function
		function t_List.Insert(_v() as integer, byval index as integer = LIST_LAST) as integer
			index = this.__ParseIndex(index)
			this.Alloc(this._Length + (ubound(_v) - lbound(_v)))
			for i as integer = lbound(_v) to ubound(_v)
				index = this.Insert(_v(i), index) + 1
			next
			return index
		end function
		function t_List.Insert(_v() as any ptr, byval index as integer = LIST_LAST) as integer
			index = this.__ParseIndex(index)
			this.Alloc(this._Length + (ubound(_v) - lbound(_v)))
			for i as integer = lbound(_v) to ubound(_v)
				index = this.Insert(_v(i), index) + 1
			next
			return index
		end function
		function t_List.Insert(byref _p as integer, byval index as integer = LIST_LAST) as integer
			return this.Insert(cptr(any ptr, _p), index)
		end function
		function t_List.Insert(byval _p as any ptr, byval index as integer = LIST_LAST) as integer
			if this._ForceSorted then
				return This.SortedInsert(_p)
			end if
			return this.__Insert(_p, index, FALSE)
		end function
		
		
		
		
		
		function t_List.SortedInsert(byref _l as t_List) as integer
			return this.SortedInsert(@_l)
		end function
		function t_List.SortedInsert(byval _l as t_List ptr) as integer
			dim as integer killCopy
			if @this = _l then
				_l = new t_List
				*_l = this
				killCopy = TRUE
			end if
			dim as integer oldOption = this._ForceSorted
			dim as integer r
			this._ForceSorted = TRUE
			r = this.Insert(_l)
			this._ForceSorted = oldOption
			if killCopy then
				_l->destructor()
			end if
			return r
		end function
		function t_List.SortedInsert(_v() as integer) as integer
			dim as integer oldOption = this._ForceSorted
			dim as integer r
			this._ForceSorted = TRUE
			r = this.Insert(_v())
			this._ForceSorted = oldOption
			return r
		end function
		function t_List.SortedInsert(_v() as any ptr) as integer
			dim as integer oldOption = this._ForceSorted
			dim as integer r
			this._ForceSorted = TRUE
			r = this.Insert(_v())
			this._ForceSorted = oldOption
			return r
		end function
		function t_List.SortedInsert(byref _p as integer) as integer
			return this.SortedInsert(cptr(any ptr, _p))
		end function
		function t_List.SortedInsert(byval _p as any ptr) as integer
			return this.__SortedInsert(_p)
		end function
		
		function t_List.__Copy(byref _in as integer) as any ptr
			dim as integer ptr n = new integer
			*n = _in
			return n
		end function
	#endif

	#ifndef __FBLIST_PP_DISABLEMACROS
		#macro HOLDS?(__type)
			, __type
		#endmacro
		#macro EXTENSIONBEGIN
			,8741396
			#ifdef __FBLIST_PP_EXTENSIONBEGIN
				#error "Found EXTENSIONBEGIN but the previous EXTENSIONBEGIN has not been EXTENSIONEND!"
			#endif
			#define __FBLIST_PP_EXTENSIONBEGIN
		#endmacro
		#macro EXTENSIONEND
			#ifndef __FBLIST_PP_EXTENSIONBEGIN
				#error "Found EXTENSIONEND without corresponding EXTENSIONBEGIN!"
			#endif
			#undef __FBLIST_PP_EXTENSIONBEGIN
			#ifndef __FBLIST_PP_LISTISEXTENDED
				#error "Found EXTENSIONEND without EXTENSIONBEGIN!"
			#else
				end type
				__FBLIST_PP_NATIVEMETHODS
			#endif
			#undef __FBLIST_PP_LISTISEXTENDED
		#endmacro
		#macro TYPELIST?(__FBLIST_PP_LISTNAME, __FBLIST_PP_LISTTYPE, __FBLIST_EXTEND...)
			#undef __FBLIST_PP_TYPEALREADYEXISTS
			#undef __FBLIST_PP_ISOFBASICTYPE
			#undef __FBLIST_PP_TYPE_NAME
			#undef __FBLIST_PP_TYPEHASCOMPARE
			#undef __FBLIST_PP_TYPEHASCSV
			#undef __FBLIST_PP_TYPEHASCOPY
			#undef __FBLIST_PP_PFNDEFAULTCOMPARENAME
			#undef __FBLIST_PP_LISTTYPE_GLOBAL
			#undef __FBLIST_PP_LISTTYPE_STR
			#undef __FBLIST_PP_LISTISEXTENDED
			#undef __FBLIST_PP_ISNUMERIC
			#if #__FBLIST_PP_LISTTYPE="ANY"
				#error "A TYPELIST can not be 'ANY', use 'dim as t_List' instead."
			#endif
			#if typeof(__FBLIST_PP_LISTTYPE)="ANY PTR"
				#error "A TYPELIST can not be 'ANY PTR', use 'dim as t_List' instead."
			#endif
			#if typeof(__FBLIST_PP_LISTTYPE)="BYTE" or _
				typeof(__FBLIST_PP_LISTTYPE)="UBYTE" or _
				typeof(__FBLIST_PP_LISTTYPE)="SINGLE" or _
				typeof(__FBLIST_PP_LISTTYPE)="DOUBLE" or _
				typeof(__FBLIST_PP_LISTTYPE)="SHORT" or _
				typeof(__FBLIST_PP_LISTTYPE)="USHORT" or _
				typeof(__FBLIST_PP_LISTTYPE)="LONG" or _
				typeof(__FBLIST_PP_LISTTYPE)="ULONG" or _
				typeof(__FBLIST_PP_LISTTYPE)="LONGINT" or _
				typeof(__FBLIST_PP_LISTTYPE)="ULONGINT" or _
				typeof(__FBLIST_PP_LISTTYPE)="INTEGER" or _
				typeof(__FBLIST_PP_LISTTYPE)="UINTEGER"
				#define __FBLIST_PP_ISNUMERIC
			#endif
			#if typeof(__FBLIST_PP_LISTTYPE)="BYTE" or _
				typeof(__FBLIST_PP_LISTTYPE)="UBYTE" or _
				typeof(__FBLIST_PP_LISTTYPE)="SINGLE" or _
				typeof(__FBLIST_PP_LISTTYPE)="DOUBLE" or _
				typeof(__FBLIST_PP_LISTTYPE)="STRING" or _
				typeof(__FBLIST_PP_LISTTYPE)="SHORT" or _
				typeof(__FBLIST_PP_LISTTYPE)="USHORT" or _
				typeof(__FBLIST_PP_LISTTYPE)="LONG" or _
				typeof(__FBLIST_PP_LISTTYPE)="ULONG" or _
				typeof(__FBLIST_PP_LISTTYPE)="LONGINT" or _
				typeof(__FBLIST_PP_LISTTYPE)="ULONGINT" or _
				typeof(__FBLIST_PP_LISTTYPE)="INTEGER" or _
				typeof(__FBLIST_PP_LISTTYPE)="UINTEGER"
				#define __FBLIST_PP_ISOFBASICTYPE
			#endif
			#define __FBLIST_PP_LISTTYPE_GLOBAL __FBLIST_PP_LISTTYPE
			#define __FBLIST_PP_LISTTYPE_STR #__FBLIST_PP_LISTTYPE
			
			#if #__FBLIST_EXTEND="8741396"
				#define __FBLIST_PP_LISTISEXTENDED
			#endif
			
			#ifndef __FBLIST_PP_ISOFBASICTYPE
				#ifdef __FBLIST_PP_LISTTYPE.__T_LIST_CSVNAME
					#if isProcedure(__FBLIST_PP_LISTTYPE.__T_LIST_CSVNAME)
						#define __FBLIST_PP_TYPEHASCSV
					#endif
				#endif
				#ifdef __FBLIST_PP_LISTTYPE.__T_LIST_COMPARENAME
					#if isProcedure(__FBLIST_PP_LISTTYPE.__T_LIST_COMPARENAME)
						#define __FBLIST_PP_TYPEHASCOMPARE
					#endif
				#endif
				#ifdef __FBLIST_PP_LISTTYPE.__T_LIST_COPYNAME
					#if isProcedure(__FBLIST_PP_LISTTYPE.__T_LIST_COPYNAME)
						#define __FBLIST_PP_TYPEHASCOPY
					#endif
				#endif
			#endif
			
			#ifndef __FBLIST_PP_LISTISEXTENDED
				#ifdef __FBLIST_T_##__FBLIST_PP_LISTTYPE
					#define __FBLIST_PP_TYPEALREADYEXISTS
					type __FBLIST_PP_LISTNAME as __FBLIST_T_##__FBLIST_PP_LISTTYPE
				#endif
			#endif
			#ifndef __FBLIST_PP_TYPEALREADYEXISTS
				#ifndef __FBLIST_PP_LISTISEXTENDED
					#define __FBLIST_PP_TYPE_NAME __FBLIST_T_##__FBLIST_PP_LISTTYPE
					type __FBLIST_PP_LISTNAME as __FBLIST_PP_TYPE_NAME
				#else '// Extended
					#define __FBLIST_PP_TYPE_NAME __FBLIST_PP_LISTNAME
				#endif
				#define __FBLIST_PP_PFNDEFAULTCOMPARENAME __FBLIST_PP_TYPE_NAME#__fnDefaultCompare
				declare function __FBLIST_PP_PFNDEFAULTCOMPARENAME(byval __a as any ptr, byval __b as any ptr) as integer
				#ifndef FBLIST_INCLUDE_HEADER_ONLY
					function __FBLIST_PP_PFNDEFAULTCOMPARENAME(byval __a as any ptr, byval __b as any ptr) as integer
						dim as __FBLIST_PP_LISTTYPE ptr _a = __a
						dim as __FBLIST_PP_LISTTYPE ptr _b = __b
						#ifdef __FBLIST_PP_ISOFBASICTYPE
							if *_a < *_b then
								return -1
							elseif *_a > *_b then
								return 1
							end if
							return 0
						#else
							#ifdef __FBLIST_PP_TYPEHASCOMPARE
								return _a->__T_LIST_COMPARENAME(*_b)
							#else
								if _a < _b then
									return -1
								elseif _a > _b then
									return 1
								end if
								return 0
							#endif
						#endif
					end function
				#endif
				type __FBLIST_PP_TYPE_NAME extends __T_LIST_FOUNDATION
					declare function Fill(byref _l as __FBLIST_PP_TYPE_NAME, byval _len as integer, byval index as integer = LIST_LAST) as integer
					declare function Fill(byval _l as __FBLIST_PP_TYPE_NAME ptr, byval _len as integer, byval index as integer = LIST_LAST) as integer
					declare function Fill(_v() as __FBLIST_PP_LISTTYPE, byval _len as integer, byval index as integer = LIST_LAST) as integer
					declare function Fill(_v() as __FBLIST_PP_LISTTYPE ptr, byval _len as integer, byval index as integer = LIST_LAST) as integer
					declare function Fill(byref _p as __FBLIST_PP_LISTTYPE, byval _len as integer, byval index as integer = LIST_LAST) as integer
					declare function Fill(byval _p as __FBLIST_PP_LISTTYPE ptr, byval _len as integer, byval index as integer = LIST_LAST) as integer
					
					#ifdef __FBLIST_PP_ISNUMERIC
						declare function IntSequence(byval _len as integer, byval index as integer = LIST_LAST) as __FBLIST_PP_LISTTYPE_GLOBAL
						declare function PoTSequence(byval _len as integer, byval index as integer = LIST_LAST) as __FBLIST_PP_LISTTYPE_GLOBAL
						declare function OddSequence(byval _len as integer, byval index as integer = LIST_LAST) as __FBLIST_PP_LISTTYPE_GLOBAL
						declare function EvenSequence(byval _len as integer, byval index as integer = LIST_LAST) as __FBLIST_PP_LISTTYPE_GLOBAL
						declare sub ListIsNumeric()
					#endif
					declare function Sequence(byval pfnGenerator as function(byval listIndex as integer, byval sequenceCount as integer) as __FBLIST_PP_LISTTYPE_GLOBAL, byval _len as integer, byval index as integer = LIST_LAST) as integer
					
					declare function Merge(byref _l as __FBLIST_PP_TYPE_NAME, byval index as integer) as integer
					declare function Merge(byval _l as __FBLIST_PP_TYPE_NAME ptr, byval index as integer) as integer
					declare function Merge(_v() as __FBLIST_PP_LISTTYPE, byval index as integer) as integer
					declare function Merge(_v() as __FBLIST_PP_LISTTYPE ptr, byval index as integer) as integer
					declare function Merge(byref _p as __FBLIST_PP_LISTTYPE, byval index as integer) as integer
					declare function Merge(byval _p as __FBLIST_PP_LISTTYPE ptr, byval index as integer) as integer
					
					declare function Replace(byref _l as __FBLIST_PP_TYPE_NAME, byval index as integer) as integer
					declare function Replace(byval _l as __FBLIST_PP_TYPE_NAME ptr, byval index as integer) as integer
					declare function Replace(_v() as __FBLIST_PP_LISTTYPE, byval index as integer) as integer
					declare function Replace(_v() as __FBLIST_PP_LISTTYPE ptr, byval index as integer) as integer
					declare function Replace(byref _p as __FBLIST_PP_LISTTYPE, byval index as integer) as integer
					declare function Replace(byval _p as __FBLIST_PP_LISTTYPE ptr, byval index as integer) as integer
					
					declare function Insert(byref _l as __FBLIST_PP_TYPE_NAME, byval index as integer = LIST_LAST) as integer
					declare function Insert(byval _l as __FBLIST_PP_TYPE_NAME ptr, byval index as integer = LIST_LAST) as integer
					declare function Insert(_v() as __FBLIST_PP_LISTTYPE, byval index as integer = LIST_LAST) as integer
					declare function Insert(_v() as __FBLIST_PP_LISTTYPE ptr, byval index as integer = LIST_LAST) as integer
					declare function Insert(byref _p as __FBLIST_PP_LISTTYPE, byval index as integer = LIST_LAST) as integer
					declare function Insert(byval _p as __FBLIST_PP_LISTTYPE ptr, byval index as integer = LIST_LAST) as integer
					
					#if 0 '// Not supported by FBC
						declare operator +=(_v() as __FBLIST_PP_LISTTYPE)
						declare operator +=(_v() as __FBLIST_PP_LISTTYPE ptr)
					#endif
					declare operator +=(byref _l as __FBLIST_PP_TYPE_NAME)
					declare operator +=(byval _l as __FBLIST_PP_TYPE_NAME ptr)
					declare operator +=(byref _p as __FBLIST_PP_LISTTYPE)
					declare operator +=(byval _p as __FBLIST_PP_LISTTYPE ptr)
					
					
					#if 0 '// Not supported by FBC
						declare operator Let(_v() as __FBLIST_PP_LISTTYPE)
						declare operator Let(_v() as __FBLIST_PP_LISTTYPE ptr)
					#endif
					declare operator Let(byref _l as __FBLIST_PP_TYPE_NAME)
					declare operator Let(byval _l as __FBLIST_PP_TYPE_NAME ptr)
					declare operator Let(byref _p as __FBLIST_PP_LISTTYPE)
					declare operator Let(byval _p as __FBLIST_PP_LISTTYPE ptr)
					
					declare function SortedInsert(byref _l as __FBLIST_PP_TYPE_NAME) as integer
					declare function SortedInsert(byval _l as __FBLIST_PP_TYPE_NAME ptr) as integer
					declare function SortedInsert(_v() as __FBLIST_PP_LISTTYPE) as integer
					declare function SortedInsert(_v() as __FBLIST_PP_LISTTYPE ptr) as integer
					declare function SortedInsert(byref _p as __FBLIST_PP_LISTTYPE) as integer
					declare function SortedInsert(byval _p as __FBLIST_PP_LISTTYPE ptr) as integer
					
					declare Constructor (byref _l as __FBLIST_PP_TYPE_NAME)
					declare Constructor (byval _l as __FBLIST_PP_TYPE_NAME ptr)
					declare Constructor (_v() as __FBLIST_PP_LISTTYPE)
					declare Constructor (_v() as __FBLIST_PP_LISTTYPE ptr)
					declare Constructor (byref _p as __FBLIST_PP_LISTTYPE)
					declare Constructor (byval _p as __FBLIST_PP_LISTTYPE ptr)
					declare Constructor ()
					
					declare sub Clear(byref _l as __FBLIST_PP_TYPE_NAME)
					declare sub Clear(byval _l as __FBLIST_PP_TYPE_NAME ptr)
					declare sub Clear(_v() as __FBLIST_PP_LISTTYPE)
					declare sub Clear(_v() as __FBLIST_PP_LISTTYPE ptr)
					declare sub Clear(byref _p as __FBLIST_PP_LISTTYPE)
					declare sub Clear(byval _p as __FBLIST_PP_LISTTYPE ptr)
					declare sub Clear()
					
					declare function Find(byval _match as __FBLIST_PP_LISTTYPE, byval _matchIndex as integer = 0) as integer
					declare function Find(byval _match as __FBLIST_PP_LISTTYPE ptr, byval _matchIndex as integer = 0) as integer
					
					declare function GetPtr(byval index as integer) as __FBLIST_PP_LISTTYPE ptr
					
					declare operator [] (byval index as integer) byref as __FBLIST_PP_LISTTYPE
					
					declare function __T_LIST_CSVNAME() as string
					
					
					declare function __Copy(byref _in as __FBLIST_PP_LISTTYPE) as __FBLIST_PP_LISTTYPE ptr
					declare sub __eldtor(byval index as integer)
				#ifndef __FBLIST_PP_LISTISEXTENDED
					end type
					__FBLIST_PP_NATIVEMETHODS
				#endif
			#endif
		#endmacro
		#macro __FBLIST_PP_NATIVEMETHODS
			#ifndef FBLIST_INCLUDE_HEADER_ONLY
				Constructor __FBLIST_PP_TYPE_NAME(byref _l as __FBLIST_PP_TYPE_NAME)
					this.Constructor()
					this.Insert(_l)
				end constructor
				Constructor __FBLIST_PP_TYPE_NAME(byval _l as __FBLIST_PP_TYPE_NAME ptr)
					this.Constructor()
					this.Insert(_l)
				end constructor
				Constructor __FBLIST_PP_TYPE_NAME(_v() as __FBLIST_PP_LISTTYPE_GLOBAL)
					this.Constructor()
					this.Insert(_v())
				end constructor
				Constructor __FBLIST_PP_TYPE_NAME(_v() as __FBLIST_PP_LISTTYPE_GLOBAL ptr)
					this.Constructor()
					this.Insert(_v())
				end constructor
				Constructor __FBLIST_PP_TYPE_NAME(byref _p as __FBLIST_PP_LISTTYPE_GLOBAL)
					this.Constructor()
					this.Insert(_p)
				end constructor
				Constructor __FBLIST_PP_TYPE_NAME(byval _p as __FBLIST_PP_LISTTYPE_GLOBAL ptr)
					this.Constructor()
					this.Insert(_p)
				end constructor
				Constructor __FBLIST_PP_TYPE_NAME()
					base.constructor()
					this.__pfnCompare	=	@__FBLIST_PP_PFNDEFAULTCOMPARENAME
				end Constructor
				
				
				
				
				
				operator __FBLIST_PP_TYPE_NAME.[] (byval index as integer) byref as __FBLIST_PP_LISTTYPE_GLOBAL
					index = this.__ParseIndex(index)
					return *cptr(__FBLIST_PP_LISTTYPE_GLOBAL ptr, this.RawListMemory[index])
				end operator
				function __FBLIST_PP_TYPE_NAME.GetPtr(byval index as integer) as __FBLIST_PP_LISTTYPE_GLOBAL ptr
					index = this.__ParseIndex(index)
					return this.RawListMemory[index]
				end function
				
				
				
				
				
				sub __FBLIST_PP_TYPE_NAME.Clear(byref _l as __FBLIST_PP_TYPE_NAME)
					this.Clear()
					this.Insert(_l)
				end sub
				sub __FBLIST_PP_TYPE_NAME.Clear(byval _l as __FBLIST_PP_TYPE_NAME ptr)
					this.Clear()
					this.Insert(_l)
				end sub
				sub __FBLIST_PP_TYPE_NAME.Clear(_v() as __FBLIST_PP_LISTTYPE_GLOBAL)
					this.Clear()
					this.Insert(_v())
				end sub
				sub __FBLIST_PP_TYPE_NAME.Clear(_v() as __FBLIST_PP_LISTTYPE_GLOBAL ptr)
					this.Clear()
					this.Insert(_v())
				end sub
				sub __FBLIST_PP_TYPE_NAME.Clear(byref _p as __FBLIST_PP_LISTTYPE_GLOBAL)
					this.Clear()
					this.Insert(_p)
				end sub
				sub __FBLIST_PP_TYPE_NAME.Clear(byval _p as __FBLIST_PP_LISTTYPE_GLOBAL ptr)
					this.Clear()
					this.Insert(_p)
				end sub
				sub __FBLIST_PP_TYPE_NAME.Clear()
					base.Clear()
				end sub
				
				
				
				
				
				#if 0  '// Not supported by FBC
					operator __FBLIST_PP_TYPE_NAME.Let(_v() as __FBLIST_PP_LISTTYPE)
						this.Clear(_v())
					end operator
					operator __FBLIST_PP_TYPE_NAME.Let(_v() as __FBLIST_PP_LISTTYPE ptr)
						this.Clear(_v())
					end operator
				#endif
				operator __FBLIST_PP_TYPE_NAME.Let(byref _l as __FBLIST_PP_TYPE_NAME)
					this.Clear(_l)
				end operator
				operator __FBLIST_PP_TYPE_NAME.Let(byval _l as __FBLIST_PP_TYPE_NAME ptr)
					this.Clear(_l)
				end operator
				operator __FBLIST_PP_TYPE_NAME.Let(byref _p as __FBLIST_PP_LISTTYPE_GLOBAL)
					this.Clear(_p)
				end operator
				operator __FBLIST_PP_TYPE_NAME.Let(byval _p as __FBLIST_PP_LISTTYPE_GLOBAL ptr)
					this.Clear(_p)
				end operator
				
				
				
				
				function __FBLIST_PP_TYPE_NAME.Fill(byref _l as __FBLIST_PP_TYPE_NAME, byval _len as integer, byval index as integer) as integer
					return this.Fill(@_l, _len, index)
				end function
				function __FBLIST_PP_TYPE_NAME.Fill(byval _l as __FBLIST_PP_TYPE_NAME ptr, byval _len as integer, byval index as integer) as integer
					dim as integer killCopy
					if @this = _l then
						_l = new __FBLIST_PP_TYPE_NAME
						*_l = this
						killCopy = TRUE
					end if
					if _l->_Length > 0 then
						index = this.__ParseIndex(index)
						for g as integer = 0 to _len -1
							index = this.Insert(cptr(__FBLIST_PP_LISTTYPE_GLOBAL ptr, _l->RawListMemory[(g mod _l->_Length)]), index) + 1
						next
					end if
					if killCopy then
						_l->Destructor
					end if
					return index
				end function
				function __FBLIST_PP_TYPE_NAME.Fill(_v() as __FBLIST_PP_LISTTYPE_GLOBAL, byval _len as integer, byval index as integer) as integer
					index = this.__ParseIndex(index)
					dim as integer lb = lbound(_v)
					dim as integer al = ubound(_v) - lb + 1
					for g as integer = 0 to _len -1
						index = this.Insert(_v(lb + (g mod al)), index) + 1
					next
					return index
				end function
				function __FBLIST_PP_TYPE_NAME.Fill(_v() as __FBLIST_PP_LISTTYPE_GLOBAL ptr, byval _len as integer, byval index as integer) as integer
					index = this.__ParseIndex(index)
					dim as integer lb = lbound(_v)
					dim as integer al = ubound(_v) - lb + 1
					for g as integer = 0 to _len -1
						index = this.Insert(_v(lb + (g mod al)), index) + 1
					next
					return index
				end function
				function __FBLIST_PP_TYPE_NAME.Fill(byref _p as __FBLIST_PP_LISTTYPE_GLOBAL, byval _len as integer, byval index as integer) as integer
					index = this.__ParseIndex(index)
					for g as integer = 0 to _len -1
						index = this.Insert(_p, index) + 1
					next
					return index
				end function
				function __FBLIST_PP_TYPE_NAME.Fill(byval _p as __FBLIST_PP_LISTTYPE_GLOBAL ptr, byval _len as integer, byval index as integer) as integer
					index = this.__ParseIndex(index)
					for g as integer = 0 to _len -1
						index = this.Insert(_p, index) + 1
					next
					return index
				end function
				
				
				#ifdef __FBLIST_PP_ISNUMERIC
					function __FBLIST_PP_TYPE_NAME.IntSequence(byval _len as integer, byval index as integer) as __FBLIST_PP_LISTTYPE_GLOBAL
						index = this.__ParseIndex(index)
						for g as integer = 0 to _len -1
							index = this.Insert(new __FBLIST_PP_LISTTYPE_GLOBAL, index)
							if index > -1 then
								this[index]	=	FBLIST_INTSEQUENCE(index, g)
							end if
							index += 1
						next
						return index
					end function
					function __FBLIST_PP_TYPE_NAME.PoTSequence(byval _len as integer, byval index as integer) as __FBLIST_PP_LISTTYPE_GLOBAL
						index = this.__ParseIndex(index)
						for g as integer = 0 to _len -1
							index = this.Insert(new __FBLIST_PP_LISTTYPE_GLOBAL, index)
							if index > -1 then
								this[index]	=	FBLIST_POTSEQUENCE(index, g)
							end if
							index += 1
						next
						return index
					end function
					function __FBLIST_PP_TYPE_NAME.EvenSequence(byval _len as integer, byval index as integer) as __FBLIST_PP_LISTTYPE_GLOBAL
						index = this.__ParseIndex(index)
						for g as integer = 0 to _len -1
							index = this.Insert(new __FBLIST_PP_LISTTYPE_GLOBAL, index)
							if index > -1 then
								this[index]	=	FBLIST_EVENSEQUENCE(index, g)
							end if
							index += 1
						next
						return index
					end function
					function __FBLIST_PP_TYPE_NAME.OddSequence(byval _len as integer, byval index as integer) as __FBLIST_PP_LISTTYPE_GLOBAL
						index = this.__ParseIndex(index)
						for g as integer = 0 to _len -1
							index = this.Insert(new __FBLIST_PP_LISTTYPE_GLOBAL, index)
							if index > -1 then
								this[index]	=	FBLIST_ODDSEQUENCE(index, g)
							end if
							index += 1
						next
						return index
					end function
				#endif
				function __FBLIST_PP_TYPE_NAME.Sequence(byval pfnGenerator as function(byval listIndex as integer, byval sequenceCount as integer) as __FBLIST_PP_LISTTYPE_GLOBAL, byval _len as integer, byval index as integer) as integer
					if pfnGenerator then
						index = this.__ParseIndex(index)
						for g as integer = 0 to _len -1
							index = this.Insert(new __FBLIST_PP_LISTTYPE_GLOBAL, index)
							if index > -1 then
								this[index]	=	pfnGenerator(index, g)
							end if
							index += 1
						next
					end if
					return index
				end function
				
				
				
				function __FBLIST_PP_TYPE_NAME.Merge(byref __l as __FBLIST_PP_TYPE_NAME, byval index as integer) as integer
					dim as __FBLIST_PP_TYPE_NAME ptr _l = @__l
					dim as integer killCopy
					if @this = _l then
						_l = new __FBLIST_PP_TYPE_NAME
						*_l = this
						killCopy = TRUE
					end if
					dim as integer ni = this.__ParseIndex(index)
					dim as integer ri = -1
					for g as integer = 0 to _l->_Length - 1
						ri = this.Merge( *cptr(__FBLIST_PP_LISTTYPE_GLOBAL ptr, _l->RawListMemory[g]), ni)
						if ri > -1 then
							ni += 1
						else
							exit for
						end if
					next
					if killCopy then
						delete _l
					end if
					return ri
				end function
				function __FBLIST_PP_TYPE_NAME.Merge(byval __l as __FBLIST_PP_TYPE_NAME ptr, byval index as integer) as integer
					dim as __FBLIST_PP_TYPE_NAME ptr _l = __l
					dim as integer killCopy
					if @this = _l then
						_l = new __FBLIST_PP_TYPE_NAME
						*_l = this
					end if
					dim as integer ni = this.__ParseIndex(index)
					dim as integer ri = -1
					for g as integer = 0 to _l->_Length - 1
						ri = this.Merge( cptr(__FBLIST_PP_LISTTYPE_GLOBAL ptr, _l->RawListMemory[g]), ni)
						if ri > -1 then
							ni += 1
						else
							exit for
						end if
					next
					if killCopy then
						delete _l
					end if
					return ri
				end function
				function __FBLIST_PP_TYPE_NAME.Merge(_v() as __FBLIST_PP_LISTTYPE_GLOBAL, byval index as integer) as integer
					dim as integer ni = this.__ParseIndex(index)
					dim as integer ri = -1
					for g as integer = lbound(_v) to ubound(_v)
						ri = this.Merge( _v(g), ni)
						if ri > -1 then
							ni += 1
						else
							exit for
						end if
					next
					return ri
				end function
				function __FBLIST_PP_TYPE_NAME.Merge(_v() as __FBLIST_PP_LISTTYPE_GLOBAL ptr, byval index as integer) as integer
					dim as integer ni = this.__ParseIndex(index)
					dim as integer ri = -1
					for g as integer = lbound(_v) to ubound(_v)
						ri = this.Merge( _v(g), ni)
						if ri > -1 then
							ni += 1
						else
							exit for
						end if
					next
					return ri
				end function
				function __FBLIST_PP_TYPE_NAME.Merge(byref _p as __FBLIST_PP_LISTTYPE_GLOBAL, byval index as integer) as integer
					return this.Merge(this.__Copy(_p), index)
				end function
				function __FBLIST_PP_TYPE_NAME.Merge(byval _p as __FBLIST_PP_LISTTYPE_GLOBAL ptr, byval index as integer) as integer
					return this.__Merge(_p, index)
				end function
				
				
				
				
				
				function __FBLIST_PP_TYPE_NAME.Replace(byref __l as __FBLIST_PP_TYPE_NAME, byval index as integer) as integer
					dim as __FBLIST_PP_TYPE_NAME ptr _l = @__l
					dim as integer killCopy
					if @this = _l then
						_l = new __FBLIST_PP_TYPE_NAME
						*_l = this
						killCopy = TRUE
					end if
					dim as integer ni = this.__ParseIndex(index)
					dim as integer ri = -1
					for g as integer = 0 to _l->_Length - 1
						if ni < this._Length then
							ri = this.Replace( *cptr(__FBLIST_PP_LISTTYPE_GLOBAL ptr, _l->RawListMemory[g]), ni)
							if ri > -1 then
								ni += 1
							else
								exit for
							end if
						else
							exit for
						end if
					next
					if killCopy then
						delete _l
					end if
					return ri
				end function
				function __FBLIST_PP_TYPE_NAME.Replace(byval __l as __FBLIST_PP_TYPE_NAME ptr, byval index as integer) as integer
					dim as __FBLIST_PP_TYPE_NAME ptr _l = __l
					dim as integer killCopy
					if @this = _l then
						_l = new __FBLIST_PP_TYPE_NAME
						*_l = this
						killCopy = TRUE
					end if
					dim as integer ni = this.__ParseIndex(index)
					dim as integer ri = -1
					for g as integer = 0 to _l->_Length - 1
						if ni < this._Length then
							ri = this.Replace( cptr(__FBLIST_PP_LISTTYPE_GLOBAL ptr, _l->RawListMemory[g]), ni)
							if ri > -1 then
								ni += 1
							else
								exit for
							end if
						else
							exit for
						end if
					next
					if killCopy then
						delete _l
					end if
					return ri
				end function
				function __FBLIST_PP_TYPE_NAME.Replace(_v() as __FBLIST_PP_LISTTYPE_GLOBAL, byval index as integer) as integer
					dim as integer ni = this.__ParseIndex(index)
					dim as integer ri = -1
					for g as integer = lbound(_v) to ubound(_v)
						if ni < this._Length then
							ri = this.Replace( _v(g), ni)
							if ri > -1 then
								ni += 1
							else
								exit for
							end if
						else
							exit for
						end if
					next
					return ri
				end function
				function __FBLIST_PP_TYPE_NAME.Replace(_v() as __FBLIST_PP_LISTTYPE_GLOBAL ptr, byval index as integer) as integer
					dim as integer ni = this.__ParseIndex(index)
					dim as integer ri = -1
					for g as integer = lbound(_v) to ubound(_v)
						if ni < this._Length then
							ri = this.Replace( _v(g), ni)
							if ri > -1 then
								ni += 1
							else
								exit for
							end if
						else
							exit for
						end if
					next
					return ri
				end function
				function __FBLIST_PP_TYPE_NAME.Replace(byref _p as __FBLIST_PP_LISTTYPE_GLOBAL, byval index as integer) as integer
					return this.Replace(this.__Copy(_p), index)
				end function
				function __FBLIST_PP_TYPE_NAME.Replace(byval _p as __FBLIST_PP_LISTTYPE_GLOBAL ptr, byval index as integer) as integer
					return this.__Replace(_p, index)
				end function
				
				
				
				
				
				operator __FBLIST_PP_TYPE_NAME.+=(byref _l as __FBLIST_PP_TYPE_NAME)
					this.Insert(_l)
				end operator
				operator __FBLIST_PP_TYPE_NAME.+=(byval _l as __FBLIST_PP_TYPE_NAME ptr)
					this.Insert(_l)
				end operator
				operator __FBLIST_PP_TYPE_NAME.+=(byref _p as __FBLIST_PP_LISTTYPE_GLOBAL)
					this.Insert(_p)
				end operator
				operator __FBLIST_PP_TYPE_NAME.+=(byval _p as __FBLIST_PP_LISTTYPE_GLOBAL ptr)
					this.Insert(_p)
				end operator
				
				
				
				
				
				function __FBLIST_PP_TYPE_NAME.Find(byval _match as __FBLIST_PP_LISTTYPE_GLOBAL, byval _matchIndex as integer = 0) as integer
					return this.__Find(@_match, _matchIndex)
				end function
				function __FBLIST_PP_TYPE_NAME.Find(byval _match as __FBLIST_PP_LISTTYPE_GLOBAL ptr, byval _matchIndex as integer = 0) as integer
					return this.__Find(_match, _matchIndex)
				end function
				
				
				
				
				
				function __FBLIST_PP_TYPE_NAME.Insert(byref __l as __FBLIST_PP_TYPE_NAME, byval index as integer = LIST_LAST) as integer
					dim as __FBLIST_PP_TYPE_NAME ptr _l = @__l
					dim as integer killCopy
					if @this = _l then
						_l = new __FBLIST_PP_TYPE_NAME
						*_l = this
						killCopy = TRUE
					end if
					index = this.__ParseIndex(index)
					this.Alloc(this._Length + _l->_Length)
					for i as integer = 0 to _l->_Length - 1
						index = this.insert(*cptr(__FBLIST_PP_LISTTYPE_GLOBAL ptr, _l->RawListMemory[i]), index) + 1
					next
					if killCopy then
						delete _l
					end if
					return index
				end function
				function __FBLIST_PP_TYPE_NAME.Insert(byval __l as __FBLIST_PP_TYPE_NAME ptr, byval index as integer = LIST_LAST) as integer
					dim as __FBLIST_PP_TYPE_NAME ptr _l = __l
					dim as integer killCopy
					if @this = _l then
						_l = new __FBLIST_PP_TYPE_NAME
						*_l = this
						killCopy = TRUE
					end if
					index = this.__ParseIndex(index)
					this.Alloc(this._Length + _l->_Length)
					for i as integer = 0 to _l->_Length - 1
						index = this.Insert(cptr(__FBLIST_PP_LISTTYPE_GLOBAL ptr, _l->RawListMemory[i]), index) + 1
					next
					if killCopy then
						delete _l
					end if
					return index
				end function
				function __FBLIST_PP_TYPE_NAME.Insert(_v() as __FBLIST_PP_LISTTYPE_GLOBAL, byval index as integer = LIST_LAST) as integer
					index = this.__ParseIndex(index)
					this.Alloc(this._Length + (ubound(_v) - lbound(_v)))
					for i as integer = lbound(_v) to ubound(_v)
						index = this.Insert(_v(i), index) + 1
					next
					return index
				end function
				function __FBLIST_PP_TYPE_NAME.Insert(_v() as __FBLIST_PP_LISTTYPE_GLOBAL ptr, byval index as integer = LIST_LAST) as integer
					index = this.__ParseIndex(index)
					this.Alloc(this._Length + (ubound(_v) - lbound(_v)))
					for i as integer = lbound(_v) to ubound(_v)
						index = this.Insert(_v(i), index) + 1
					next
					return index
				end function
				function __FBLIST_PP_TYPE_NAME.Insert(byref _p as __FBLIST_PP_LISTTYPE_GLOBAL, byval index as integer = LIST_LAST) as integer
					return this.Insert(this.__Copy(_p), index)
				end function
				function __FBLIST_PP_TYPE_NAME.Insert(byval _p as __FBLIST_PP_LISTTYPE_GLOBAL ptr, byval index as integer = LIST_LAST) as integer
					if this._ForceSorted then
						return This.SortedInsert(_p)
					end if
					return this.__Insert(_p, index, FALSE)
				end function
				
				
				
				
				
				function __FBLIST_PP_TYPE_NAME.SortedInsert(byref __l as __FBLIST_PP_TYPE_NAME) as integer
					dim as __FBLIST_PP_TYPE_NAME ptr _l = @__l
					dim as integer killCopy
					if @this = _l then
						_l = new __FBLIST_PP_TYPE_NAME
						*_l = this
						killCopy = TRUE
					end if
					dim as integer oldOption = this._ForceSorted
					dim as integer r
					this._ForceSorted = TRUE
					r = this.Insert(_l)
					this._ForceSorted = oldOption
					if killCopy then
						delete _l
					end if
					return r
				end function
				function __FBLIST_PP_TYPE_NAME.SortedInsert(byval __l as __FBLIST_PP_TYPE_NAME ptr) as integer
					dim as __FBLIST_PP_TYPE_NAME ptr _l = __l
					dim as integer killCopy
					if @this = _l then
						_l = new __FBLIST_PP_TYPE_NAME
						*_l = this
						killCopy = TRUE
					end if
					dim as integer oldOption = this._ForceSorted
					dim as integer r
					this._ForceSorted = TRUE
					r = this.Insert(_l)
					this._ForceSorted = oldOption
					if killCopy then
						delete _l
					end if
					return r
				end function
				function __FBLIST_PP_TYPE_NAME.SortedInsert(_v() as __FBLIST_PP_LISTTYPE_GLOBAL) as integer
					dim as integer oldOption = this._ForceSorted
					dim as integer r
					this._ForceSorted = TRUE
					r = this.Insert(_v())
					this._ForceSorted = oldOption
					return r
				end function
				function __FBLIST_PP_TYPE_NAME.SortedInsert(_v() as __FBLIST_PP_LISTTYPE_GLOBAL ptr) as integer
					dim as integer oldOption = this._ForceSorted
					dim as integer r
					this._ForceSorted = TRUE
					r = this.Insert(_v())
					this._ForceSorted = oldOption
					return r
				end function
				function __FBLIST_PP_TYPE_NAME.SortedInsert(byref _p as __FBLIST_PP_LISTTYPE_GLOBAL) as integer
					return this.SortedInsert(this.__Copy(_p))
				end function
				function __FBLIST_PP_TYPE_NAME.SortedInsert(byval _p as __FBLIST_PP_LISTTYPE_GLOBAL ptr) as integer
					return this.__SortedInsert(_p)
				end function
				
				
				
				
				
				function __FBLIST_PP_TYPE_NAME.__T_LIST_CSVNAME() as string
					dim as string r
					if this._Length then
						dim as __FBLIST_PP_LISTTYPE_GLOBAL ptr n
						for i as integer = 0 to this._Length - 1
							n = this.RawListMemory[i]
							if i > 0 then
								r &=  ", "
							end if
							#ifdef __FBLIST_PP_ISOFBASICTYPE
								r &= *n
							#else
								#ifdef __FBLIST_PP_TYPEHASCSV
									r &= n->__T_LIST_CSVNAME()
								#else
									r &= "0x"& hex(n, ___T_LIST_ENTRYSIZE*2)
								#endif
							#endif
						next
					end if
					return r
				end function
				
				
				
				
				
				function __FBLIST_PP_TYPE_NAME.__Copy(byref _in as __FBLIST_PP_LISTTYPE_GLOBAL) as __FBLIST_PP_LISTTYPE_GLOBAL ptr
					dim as __FBLIST_PP_LISTTYPE_GLOBAL ptr n
					if this._UseCopy then
						#ifdef __FBLIST_PP_TYPEHASCOPY
							n = _in.__T_LIST_COPYNAME()
						#else
							n = new __FBLIST_PP_LISTTYPE_GLOBAL
							*n = _in
						#endif
					else
						n = new __FBLIST_PP_LISTTYPE_GLOBAL
						*n = _in
					end if
					return n
				end function
				
				sub __FBLIST_PP_TYPE_NAME.__eldtor(byval index as integer)
					if index >= 0 and index < this._AllocSize then
						if this.RawListMemory[index] then
							#if __FBLIST_PP_LISTTYPE_STR<>"ANY"
								if this._UseDestructor then
									delete cptr(__FBLIST_PP_LISTTYPE_GLOBAL ptr, this.RawListMemory[index])
								end if
							#endif
							this.RawListMemory[index] = 0
						end if
					end if
				end sub
				
			#endif
		#endmacro
	#endif
#endif